@@ -12,7 +12,7 @@ from . import _apt
 from . import _flatpak
 from ._flatpak import FlatpakRemoteInfo
 from .pkgInfo import FlatpakPkgInfo, AptPkgInfo
-from .misc import print_timing
+from .misc import print_timing, debug, warn
 
 SYS_CACHE_PATH = "/var/cache/mintinstall/pkginfo.json"
 USER_CACHE_PATH = os.path.join(GLib.get_user_cache_dir(), "mintinstall", "pkginfo.json")
@@ -20,7 +20,7 @@ USER_CACHE_PATH = os.path.join(GLib.get_user_cache_dir(), "mintinstall", "pkginf
 MAX_AGE = 7 * (60 * 60 * 24) # days
 
 class CacheLoadingException(Exception):
-    '''Thrown when there was an issue loading the pickled package set'''
+    """Thrown when there was an issue loading the pickled package set"""
 
 class JsonObject(object):
     def __init__(self, pkginfo_cache, section_lists, flatpak_remote_infos):
@@ -58,11 +58,18 @@ class PkgCache(object):
     STATUS_OK = 1
 
     @print_timing
-    def __init__(self, have_flatpak=False):
+    def __init__(self, pkg_type, cache_path=None, have_flatpak=True):
         super(PkgCache, self).__init__()
 
         self.status = self.STATUS_EMPTY
-        self.have_flatpak = have_flatpak
+        self.cache_content = pkg_type
+
+        if cache_path != None:
+            self.custom_cache_path = Path(cache_path)
+        else:
+            self.custom_cache_path = None
+
+        self.have_flatpak = have_flatpak and pkg_type in ("f", None)
 
         self._items = {}
         self._item_lock = threading.Lock()
@@ -122,19 +129,25 @@ class PkgCache(object):
         sections = {}
         flatpak_remote_infos = {}
 
-        if self.have_flatpak:
+        # If there's no cache, always generate both package types.
+        if self.have_flatpak and (self.cache_content in ("f", None) or self.status == self.STATUS_EMPTY):
             cache, flatpak_remote_infos = _flatpak.process_full_flatpak_installation(cache)
 
-        cache, sections = _apt.process_full_apt_cache(cache)
+        if self.cache_content in ("a", None) or self.status == self.STATUS_EMPTY:
+            cache, sections = _apt.process_full_apt_cache(cache)
 
         return cache, sections, flatpak_remote_infos
 
     def _get_best_load_path(self):
+        # If a custom path is set, always regenerate the cache.
+        if self.custom_cache_path != None:
+            return None 
+
         try:
             sys_mtime = os.path.getmtime(SYS_CACHE_PATH)
 
             if ((time.time() - MAX_AGE) > sys_mtime) or not os.access(SYS_CACHE_PATH, os.R_OK):
-                print("Installer: System pkgcache too old or not accessible, skipping")
+                debug("Installer: System pkgcache too old or not accessible, skipping")
                 sys_mtime = 0
         except OSError:
             sys_mtime = 0
@@ -143,7 +156,7 @@ class PkgCache(object):
             user_mtime = os.path.getmtime(USER_CACHE_PATH)
 
             if (time.time() - MAX_AGE) > user_mtime:
-                print("Installer: User pkgcache too old, skipping")
+                debug("Installer: User pkgcache too old, skipping")
                 user_mtime = 0
         except OSError:
             user_mtime = 0
@@ -157,10 +170,10 @@ class PkgCache(object):
         # Select the most recent
         if sys_mtime > user_mtime:
             most_recent = SYS_CACHE_PATH
-            print("Installer: System pkgcache is most recent, using it.")
+            debug("Installer: System pkgcache is most recent, using it.")
         else:
             most_recent = USER_CACHE_PATH
-            print("Installer: User pkgcache is most recent, using it.")
+            debug("Installer: User pkgcache is most recent, using it.")
 
         return Path(most_recent)
 
@@ -187,7 +200,7 @@ class PkgCache(object):
                 sections = json_obj.section_lists
                 flatpak_remote_infos = json_obj.flatpak_remote_infos
         except Exception as e:
-            print("Installer: Error loading pkginfo cache:", e)
+            warn("Installer: Error loading pkginfo cache:", str(e))
             cache = None
 
         if cache == None:
@@ -196,6 +209,9 @@ class PkgCache(object):
         return cache, sections, flatpak_remote_infos
 
     def _get_best_save_path(self):
+        if self.custom_cache_path != None:
+            return self.custom_cache_path 
+
         best_path = None
 
         # Prefer the system location, as all users can access it
@@ -225,12 +241,24 @@ class PkgCache(object):
             with path.open(mode='w', encoding="utf8") as f:
                 json.dump(to_be_json, f, default=lambda o: o.to_json(), indent=4)
         except Exception as e:
-            print("Installer: Could not save cache:", str(e))
+            warn("Installer: Could not save cache:", str(e))
 
     def _new_cache_common(self):
-        print("Installer: Generating new pkgcache")
+        debug("Installer: Generating new pkgcache")
         cache, sections, flatpak_remote_infos = self._generate_cache()
 
+        # If we're refreshing only a specific package type, don't destroy existing
+        # items of the other type (otherwise if the cache is refreshed by mintupdate's
+        # flatpak updater mintinstall will end up starting without any apt package info
+        # and look broken).
+        with self._item_lock:
+            if self.cache_content == "f":
+                for key in [key for key in self._items.keys() if key.startswith("a")]:
+                    cache[key] = self._items[key]
+            elif self.cache_content == "a":
+                for key in [key for key in self._items.keys() if key.startswith("f")]:
+                    cache[key] = self._items[key]
+
         if len(cache) > 0:
             self._save_cache(JsonObject(cache, sections, flatpak_remote_infos))
 
@@ -262,7 +290,7 @@ class PkgCache(object):
     def force_new_cache(self):
         self._new_cache_common()
 
-    def find_pkginfo(self, string, pkg_type=None):
+    def find_pkginfo(self, string, pkg_type=None, remote=None):
         if pkg_type in (None, "a"):
             pkginfo = _apt.find_pkginfo(self, string)
 
@@ -271,8 +299,7 @@ class PkgCache(object):
 
         if self.have_flatpak:
             if pkg_type in (None, "f"):
-                pkginfo = _flatpak.find_pkginfo(self, string)
-
+                pkginfo = _flatpak.find_pkginfo(self, string, remote)
                 if pkginfo != None:
                     return pkginfo
 
@@ -292,7 +319,7 @@ class PkgCache(object):
             installer_log = gzip.open(installer_log, "r").read().decode('utf-8').splitlines()
         except Exception as e:
             # There are a number of different exceptions here, but there's only one response
-            print("Could not get initial installed packages list (check /var/log/installer/initial-status.gz): %s" % str(e))
+            warn("Could not get initial installed packages list (check /var/log/installer/initial-status.gz): %s" % str(e))
             return None
         initial_status = [x[9:] for x in installer_log if x.startswith("Package: ")]
         if not initial_status:
