@@ -36,9 +36,10 @@ class PkgInfo:
         self.summary = None
         self.description = None
         self.version = None
-        self.icon = None
+        self.icon = {}
         self.screenshots = []
-        self.url = None
+        self.homepage_url = None
+        self.help_url = None
 
         # Runtime categories
         self.categories = []
@@ -93,8 +94,6 @@ class AptPkgInfo(PkgInfo):
             summary = ""
             if candidate.summary is not None:
                 summary = candidate.summary
-                summary = summary.replace("<", "&lt;")
-                summary = summary.replace("&", "&amp;")
 
                 self.summary = capitalize(summary)
 
@@ -126,28 +125,30 @@ class AptPkgInfo(PkgInfo):
         return self.description
 
     def get_icon(self, pkginfo, apt_pkg=None, size=64):
-        if self.icon:
-            return self.icon
+        try:
+            return self.icon[size]
+        except:
+            pass
 
         # Look in app-install-data and pixmaps
         for extension in ['svg', 'png', 'xpm']:
             for suffix in ['', '-icon']:
                 icon_path = "/usr/share/app-install/icons/%s%s.%s" % (pkginfo.name, suffix, extension)
                 if os.path.exists(icon_path):
-                    self.icon = icon_path
-                    return self.icon
+                    self.icon[size] = icon_path
+                    return self.icon[size]
 
                 icon_path = "/usr/share/pixmaps/%s.%s" % (pkginfo.name, extension)
                 if os.path.exists(icon_path):
-                    self.icon = icon_path
-                    return self.icon
+                    self.icon[size] = icon_path
+                    return self.icon[size]
 
         theme = Gtk.IconTheme.get_default()
 
         for name in [pkginfo.name, pkginfo.name.split(":")[0], pkginfo.name.split("-")[0], pkginfo.name.split(".")[-1].lower()]:
             if theme.has_icon(name):
-                self.icon = name
-                return self.icon
+                self.icon[size] = name
+                return self.icon[size]
 
         return None
 
@@ -164,26 +165,29 @@ class AptPkgInfo(PkgInfo):
             else:
                 self.version = apt_pkg.candidate.version
 
-            if self.version == None:
-                self.version = ""
+        if self.version == None:
+            self.version = ""
 
         return self.version
 
-    def get_url(self, apt_pkg=None):
-        if self.url:
-            return self.url
+    def get_homepage_url(self, apt_pkg=None):
+        if self.homepage_url:
+            return self.homepage_url
 
         if apt_pkg:
             if apt_pkg.is_installed:
-                self.url = apt_pkg.installed.homepage
+                self.homepage_url = apt_pkg.installed.homepage
             else:
-                self.url = apt_pkg.candidate.homepage
+                self.homepage_url = apt_pkg.candidate.homepage
 
-        if self.url == None:
-            self.url = ""
+        if self.homepage_url == None:
+            self.homepage_url = ""
 
-        return self.url
+        return self.homepage_url
 
+    def get_help_url(self, apt_pkg=None):
+        # We can only get the homepage from apt
+        return ""
 
 class FlatpakPkgInfo(PkgInfo):
     def __init__(self, pkg_hash=None, remote=None, ref=None, remote_url=None, installed=False):
@@ -196,6 +200,8 @@ class FlatpakPkgInfo(PkgInfo):
         self.remote = remote # "flathub"
         self.remote_url = remote_url
 
+        self.installed = installed
+
         self.refid = ref.format_ref() # app/org.foo.Bar/x86_64/stable
         self.kind = ref.get_kind() # Will be app for now
         self.arch = ref.get_arch()
@@ -285,96 +291,67 @@ class FlatpakPkgInfo(PkgInfo):
         return self.description
 
     def get_icon(self, pkginfo, as_component=None, size=64):
-        if self.icon:
-            return self.icon
+        try:
+            return self.icon[size]
+        except:
+            pass
 
         if as_component:
             icons = as_component.get_icons()
 
             if icons:
                 icon_to_use = None
-                first_icon = None
+                remote_icon = None
+                first_local_icon = None
+                good_size_icon = None
 
                 for icon in icons:
-                    if first_icon == None:
-                        first_icon = icon
+                    if icon.get_kind() == AppStreamGlib.IconKind.REMOTE:
+                        remote_icon = icon
+                        continue
 
-                    if icon.get_height() == size:
-                        icon_to_use = icon
-                        break
+                    if icon.get_kind() in (AppStreamGlib.IconKind.LOCAL,  \
+                                           AppStreamGlib.IconKind.CACHED, \
+                                           AppStreamGlib.IconKind.STOCK):
+                        first_local_icon = icon
+                        if icon.get_height() == size or ("%dx%d" % (size, size)) in icon.get_prefix():
+                            good_size_icon = icon
+                            break
 
-                if icon_to_use == None:
-                    icon_to_use = first_icon
+                icon_to_use = good_size_icon or first_local_icon or remote_icon
+                kind = icon_to_use.get_kind()
 
-                if icon_to_use.get_kind() in (AppStreamGlib.IconKind.LOCAL, AppStreamGlib.IconKind.CACHED):
-                    self.icon = os.path.join(icon_to_use.get_prefix(), icon_to_use.get_name())
-                elif icon_to_use.get_kind() == AppStreamGlib.IconKind.REMOTE:
-                    self.icon = icon_to_use.get_url()
-                elif icon_to_use.get_kind() == AppStreamGlib.IconKind.STOCK:
-                    self.icon = icon_to_use.get_name()
+                if kind in (AppStreamGlib.IconKind.LOCAL, AppStreamGlib.IconKind.CACHED):
+                    self.icon[size] = os.path.join(icon_to_use.get_prefix(), icon_to_use.get_name())
+                elif kind == AppStreamGlib.IconKind.REMOTE:
+                    self.icon[size] = icon_to_use.get_url()
+                elif kind == AppStreamGlib.IconKind.STOCK:
+                    self.icon[size] = icon_to_use.get_name()
 
-        if self.icon == None:
-            self.icon == ""
+        try:
+            return self.icon[size]
+        except:
             return None
 
-        return self.icon
-
     def get_screenshots(self, as_component=None):
         if len(self.screenshots) > 0:
             return self.screenshots
 
         if as_component:
-            screenshots = as_component.get_screenshots()
-
-            for ss in screenshots:
-                images = ss.get_images()
-
-                if len(images) == 0:
-                    continue
-
-                # FIXME: there must be a better way.  Finding an optimal size to use without just
-                # resorting to an original source.
-
-                best = None
-                largest = None
-
-                for image in images:
-                    if image.get_kind() == AppStreamGlib.ImageKind.SOURCE:
-                        continue
-
-                    w = image.get_width()
-
-                    if w > 500 and w < 625:
-                        best = image
-                        break
-
-                    if w > 625:
-                        continue
-
-                    if largest == None or (largest != None and largest.get_width() < w):
-                        largest = image
-
-                if best == None and largest == None:
-                    continue
-
-                if best == None:
-                    best = largest
-
-                if ss.get_kind() == AppStreamGlib.ScreenshotKind.DEFAULT:
-                    self.screenshots.insert(0, best.get_url())
-                else:
-                    self.screenshots.append(best.get_url())
+            self.screenshots = as_component.get_screenshots()
 
         return self.screenshots
 
     def get_version(self, as_component=None):
         if self.version:
+            # as_component.get_release_default().get_version()
             return self.version
 
         if as_component:
             releases = as_component.get_releases()
 
             if len(releases) > 0:
+                releases.sort(key=lambda r: r.get_timestamp(), reverse=True)
                 version = releases[0].get_version()
 
                 if version:
@@ -385,17 +362,30 @@ class FlatpakPkgInfo(PkgInfo):
 
         return self.version
 
-    def get_url(self, as_component=None):
-        if self.url:
-            return self.url
+    def get_homepage_url(self, as_component=None):
+        if self.homepage_url:
+            return self.homepage_url
 
         if as_component:
-            url = as_component.get_url(AppStreamGlib.UrlKind.HOMEPAGE)
+            url = as_component.get_url_item(AppStreamGlib.UrlKind.HOMEPAGE)
 
             if url != None:
-                self.url = url
+                self.homepage_url = url
 
-        if self.url == None:
-            self.url = self.remote_url
+        return self.homepage_url
+
+    def get_help_url(self, as_component=None):
+        if self.help_url:
+            return self.help_url
+
+        if as_component:
+            url = as_component.get_url_item(AppStreamGlib.UrlKind.HELP)
+
+            if url != None:
+                self.help_url = url
+
+        if self.help_url == None:
+            self.help_url = ""
+
+        return self.help_url
 
-        return self.url
\ No newline at end of file
